@@ -39,22 +39,15 @@ static struct wl_list wlEglDeviceDpyList = WL_LIST_INIT(&wlEglDeviceDpyList);
 EGLBoolean wlEglIsWaylandDisplay(void *nativeDpy)
 {
 #if HAS_MINCORE
-    void *first_pointer = NULL;
-    EGLBoolean ret;
-
     if (!wlEglPointerIsDereferencable(nativeDpy)) {
         return EGL_FALSE;
     }
 
-    first_pointer = *(void **) nativeDpy;
-
-    /* wl_display is a wl_proxy, which is a wl_object.
-     * wl_object's first element points to the interfacetype.
-     */
-    ret = (first_pointer == &wl_display_interface) ? EGL_TRUE : EGL_FALSE;
-
-    return ret;
+    return wlEglCheckInterfaceType((struct wl_object *)nativeDpy,
+                                   "wl_display_interface");
 #else
+    (void)nativeDpy;
+
     /* we return EGL_TRUE in order to always assume a valid wayland
      * display is given so that we bypass all the checks that would
      * prevent any of the functions in this library to work
@@ -66,23 +59,25 @@ EGLBoolean wlEglIsWaylandDisplay(void *nativeDpy)
 
 EGLBoolean wlEglIsValidNativeDisplayExport(void *data, void *nativeDpy)
 {
-    char *val = getenv("EGL_PLATFORM");
-    (void) data;
+    EGLBoolean  checkDpy = EGL_TRUE;
+    char       *val      = getenv("EGL_PLATFORM");
+    (void)data;
 
     if (val && !strcasecmp(val, "wayland")) {
         return EGL_TRUE;
     }
-#if HAS_MINCORE
-    return wlEglIsWaylandDisplay(nativeDpy);
-#else
+
+#if !HAS_MINCORE
     /* wlEglIsWaylandDisplay always returns true if  mincore(2)
      * is not available, hence we cannot ascertain whether the
      * the nativeDpy is wayland.
      * Note: this effectively forces applications to use
      * eglGetPlatformDisplay() instead of eglGetDisplay().
      */
-    return EGL_FALSE;
+    checkDpy = EGL_FALSE;
 #endif
+
+    return (checkDpy ? wlEglIsWaylandDisplay(nativeDpy) : EGL_FALSE);
 }
 
 EGLBoolean wlEglBindDisplaysHook(void *data, EGLDisplay dpy, void *nativeDpy)
@@ -267,6 +262,54 @@ struct wl_event_queue* wlGetEventQueue(WlEglDisplay *display)
     return (evtQueue ? evtQueue->queue : NULL);
 }
 
+void wlUpdateQueueBusyStatus(WlEglDisplay *display, struct wl_event_queue *queue, EGLBoolean isBusy)
+{
+    WlEventQueue *evtQueue = NULL;
+    WlEventQueue *iter     = NULL;
+    WlEventQueue *tmp      = NULL;
+
+    if (display != NULL) {
+        assert(queue);
+
+        /* Try to find the queue from dpy queue list */
+        wl_list_for_each(iter, &display->evtQueueList, dpyLink) {
+            if (iter->queue == queue) {
+                evtQueue = iter;
+                break;
+            }
+        }
+
+        if (evtQueue) {
+            if (isBusy) {
+                evtQueue->refCount++;
+            }
+            else {
+                /* refCount should not be already 0 */
+                assert(evtQueue->refCount > 0);
+                evtQueue->refCount--;
+            }
+        }
+        /* If it isn't present in evtQueueList, check the
+         * availability in dangling queue list if the isBusy flag is FALSE.
+         * If it is present in dangling queue list and the refCnt reaches 0,
+         * invalidate and destroy */
+        else if (!isBusy) {
+            wl_list_for_each_safe(iter, tmp, &display->dangEvtQueueList, dangLink) {
+                if (iter->queue == queue) {
+                    iter->refCount--;
+                    if (!iter->refCount) {
+                        wl_event_queue_destroy(iter->queue);
+                        iter->queue = NULL;
+                        wl_list_remove(&iter->dangLink);
+                        free(iter);
+                    }
+                    break;
+                }
+            }
+        }
+    }
+}
+
 int wlEglRoundtrip(WlEglDisplay *display, struct wl_event_queue *queue)
 {
     struct wl_display *wrapper;
@@ -362,6 +405,14 @@ static EGLBoolean terminateDisplay(EGLDisplay dpy, EGLBoolean skipDestroy)
             iter->queue = NULL;
             wl_list_remove(&iter->dpyLink);
         }
+
+        /* Invalidate and destroy event queues from dangling queue list if any */
+        wl_list_for_each_safe(iter, tmp, &display->dangEvtQueueList, dangLink) {
+            wl_event_queue_destroy(iter->queue);
+            iter->queue = NULL;
+            wl_list_remove(&iter->dangLink);
+            free(iter);
+        }
     }
     if (display->ownNativeDpy) {
         wl_display_disconnect(display->nativeDpy);
@@ -510,6 +561,7 @@ EGLDisplay wlEglGetPlatformDisplayExport(void *data,
     display->ownNativeDpy = ownNativeDpy;
     display->nativeDpy    = nativeDpy;
     wl_list_init(&display->evtQueueList);
+    wl_list_init(&display->dangEvtQueueList);
 
     queue = wlGetEventQueue(display);
     if (queue == NULL) {
@@ -829,7 +881,6 @@ const char* wlEglQueryStringExport(void *data,
                                    EGLExtPlatformString name)
 {
     WlEglPlatformData *pData   = (WlEglPlatformData *)data;
-    WlEglDisplay      *display = (WlEglDisplay *)dpy;
     EGLBoolean         isEGL15 = (pData->egl.major > 1) ||
                                  ((pData->egl.major == 1) &&
                                   (pData->egl.minor >= 5));
@@ -848,34 +899,23 @@ const char* wlEglQueryStringExport(void *data,
             res = isEGL15 ? "EGL_KHR_platform_wayland EGL_EXT_platform_wayland" :
                             "EGL_EXT_platform_wayland";
         } else {
-            EGLBoolean isWlEglDpy = EGL_FALSE;
+            /*
+             * Check whether the given display supports EGLStream
+             * extensions. For Wayland support over EGLStreams, at least the
+             * following extensions must be supported by the underlying
+             * driver:
+             *
+             *  - EGL_KHR_stream
+             *  - EGL_KHR_stream_producer_eglsurface
+             *  - EGL_KHR_stream_cross_process_fd
+             */
+            const char *exts = pData->egl.queryString(dpy, EGL_EXTENSIONS);
 
-            wlExternalApiLock();
-            isWlEglDpy = wlEglIsWlEglDisplay(display);
-            wlExternalApiUnlock();
-
-            if (isWlEglDpy) {
+            if (wlEglFindExtension("EGL_KHR_stream", exts) &&
+                wlEglFindExtension("EGL_KHR_stream_producer_eglsurface", exts) &&
+                wlEglFindExtension("EGL_KHR_stream_cross_process_fd", exts)) {
                 res = "EGL_WL_bind_wayland_display "
-                      "EGL_WL_wayland_eglstream";
-            } else {
-                /*
-                 * Check whether the given display supports EGLStream
-                 * extensions. For Wayland support over EGLStreams, at least the
-                 * following extensions must be supported by the underlying
-                 * driver:
-                 *
-                 *  - EGL_KHR_stream
-                 *  - EGL_KHR_stream_producer_eglsurface
-                 *  - EGL_KHR_stream_cross_process_fd
-                 */
-                const char *exts = pData->egl.queryString(dpy, EGL_EXTENSIONS);
-
-                if (wlEglFindExtension("EGL_KHR_stream", exts) &&
-                    wlEglFindExtension("EGL_KHR_stream_producer_eglsurface", exts) &&
-                    wlEglFindExtension("EGL_KHR_stream_cross_process_fd", exts)) {
-                    res = "EGL_WL_bind_wayland_display "
-                          "EGL_WL_wayland_eglstream";
-                }
+                     "EGL_WL_wayland_eglstream";
             }
         }
         break;
