@@ -71,6 +71,8 @@ EGLBoolean wlEglIsWaylandWindowValid(struct wl_egl_window *window)
             return EGL_FALSE;
         }
     }
+    return wlEglCheckInterfaceType((struct wl_object *)surface,
+                                   "wl_surface_interface");
 #else
     /*
      * Note that dereferencing an invalid surface pointer could mean an old
@@ -78,10 +80,11 @@ EGLBoolean wlEglIsWaylandWindowValid(struct wl_egl_window *window)
      * member in wl_egl_window struct.
      */
     surface = window->surface;
-#endif
+
     /* wl_surface is a wl_proxy, which is a wl_object. wl_objects's first
      * element points to the interface type */
     return (((*(void **)surface)) == &wl_surface_interface);
+#endif
 }
 
 static void
@@ -90,10 +93,25 @@ wayland_throttleCallback(void *data,
                          uint32_t time)
 {
     WlEglSurface *surface = (WlEglSurface *)data;
+    WlEglDisplay *display = NULL;
+
     (void) time;
 
-    surface->throttleCallback = NULL;
-    wl_callback_destroy(callback);
+    wlExternalApiLock();
+
+    /* the surface could be destroyed while the lock was suspended in
+     * wlEglWaitFrameSync */
+    if (wlEglIsWlEglSurface(surface) &&
+        (surface->throttleCallback != NULL)) {
+        display = surface->wlEglDpy;
+        wlUpdateQueueBusyStatus(display, surface->throttleCbQueue, EGL_FALSE);
+        surface->throttleCbQueue = NULL;
+
+        wl_callback_destroy(callback);
+        surface->throttleCallback = NULL;
+    }
+
+    wlExternalApiUnlock();
 }
 
 static const struct wl_callback_listener throttle_listener = {
@@ -105,6 +123,11 @@ void wlEglCreateFrameSync(WlEglSurface *surface, struct wl_event_queue *queue)
     struct wl_surface *wrapper = NULL;
 
     if (surface->swapInterval > 0) {
+        /* it is assumed that the refCount is decremented for
+         * throttleCbQueue and it is set as NULL in the previous
+         * throttleCallback, if not then raise the assertion */
+        assert(surface->throttleCbQueue == NULL);
+
         wrapper = wl_proxy_create_wrapper(surface->wlSurface);
         wl_proxy_set_queue((struct wl_proxy *)wrapper, queue);
         surface->throttleCallback = wl_surface_frame(wrapper);
@@ -113,8 +136,25 @@ void wlEglCreateFrameSync(WlEglSurface *surface, struct wl_event_queue *queue)
                                      &throttle_listener, surface) == -1) {
             return;
         }
-        wl_surface_commit(surface->wlSurface);
 
+        /* After a window resize, the compositor has to be
+         * updated with the new buffer's geometry. However, we
+         * won't have updated geometry information until the
+         * underlying buffer is attached. A surface attach
+         * may be deferred to a later time in some situations
+         * (e.g. FIFO_SYNCHRONOUS + damage thread).
+         *
+         * Therefore, the surface_commit done from here would
+         * use outdated geometry information if the buffer is
+         * not attached, which would make xdg-shell fail with
+         * error. To avoid this, skip the surface commit here
+         * if the surface attach is not yet done. */
+        if (surface->ctx.isAttached) {
+            wl_surface_commit(surface->wlSurface);
+        }
+
+        surface->throttleCbQueue = queue;
+        wlUpdateQueueBusyStatus(surface->wlEglDpy, queue, EGL_TRUE);
     }
 }
 
@@ -321,7 +361,14 @@ destroy_surface_context(WlEglSurface *surface, WlEglSurfaceCtx *ctx)
         return;
     }
 
-    if (surface->throttleCallback != NULL) {
+    /* If we are destroying the current context,
+     * we are likely tearing down, and the latest
+     * framesync must be destroyed too */
+    if ((&surface->ctx == ctx) &&
+        (surface->throttleCallback != NULL)) {
+        wlUpdateQueueBusyStatus(display, surface->throttleCbQueue, EGL_FALSE);
+        surface->throttleCbQueue = NULL;
+
         wl_callback_destroy(surface->throttleCallback);
         surface->throttleCallback = NULL;
     }
@@ -334,6 +381,11 @@ destroy_surface_context(WlEglSurface *surface, WlEglSurfaceCtx *ctx)
     }
 
     if (resource) {
+        if (ctx->wlStreamResCbQueue != NULL) {
+            wlUpdateQueueBusyStatus(display, ctx->wlStreamResCbQueue, EGL_FALSE);
+            ctx->wlStreamResCbQueue = NULL;
+        }
+
         wl_buffer_destroy(resource);
     }
 }
@@ -362,15 +414,24 @@ wl_buffer_release(void *data, struct wl_buffer *buffer)
 {
     WlEglSurface *surface = (WlEglSurface*)data;
     WlEglSurfaceCtx *ctx;
+    struct wl_event_queue *queue;
 
     /* Look for the surface context for the given buffer and destroy it */
     wlExternalApiLock();
-    wl_list_for_each(ctx, &surface->oldCtxList, link) {
-        if (ctx->wlStreamResource == buffer) {
-            destroy_surface_context(surface, ctx);
-            wl_list_remove(&ctx->link);
-            free(ctx);
-            break;
+    if (wlEglIsWlEglSurface(surface)) {
+        if (surface->ctx.wlStreamResCbQueue != NULL) {
+            queue = surface->ctx.wlStreamResCbQueue;
+            wlUpdateQueueBusyStatus(surface->wlEglDpy, queue, EGL_FALSE);
+            surface->ctx.wlStreamResCbQueue = NULL;
+        }
+
+        wl_list_for_each(ctx, &surface->oldCtxList, link) {
+            if (ctx->wlStreamResource == buffer) {
+                destroy_surface_context(surface, ctx);
+                wl_list_remove(&ctx->link);
+                free(ctx);
+                break;
+            }
         }
     }
     wlExternalApiUnlock();
@@ -424,6 +485,9 @@ create_wl_eglstream(WlEglSurface *surface,
         return NULL;
     }
 
+    surface->ctx.wlStreamResCbQueue = queue;
+    wlUpdateQueueBusyStatus(display, queue, EGL_TRUE);
+
     return buffer;
 }
 
@@ -510,7 +574,7 @@ static EGLint startInetHandshake(pthread_t *thread,
                                  int *clientSocket,
                                  int *port)
 {
-    struct sockaddr_in addr = { 0 };
+    struct sockaddr_in addr;
     unsigned int       addrLen;
     EGLint err = EGL_SUCCESS;
     int    ret;
@@ -524,6 +588,8 @@ static EGLint startInetHandshake(pthread_t *thread,
         goto fail;
     }
 
+    memset(&addr, 0, sizeof(addr));
+
     addr.sin_family      = AF_INET;
     addr.sin_addr.s_addr = INADDR_ANY; /* Accept connections to all IPs */
     addr.sin_port        = htons(0);   /* Get a random port */
@@ -766,6 +832,11 @@ create_surface_context(WlEglSurface *surface)
 
     assert(surface->ctx.eglSurface == EGL_NO_SURFACE);
 
+    /* it is assumed that the wlStreamResCbQueue is NULL and
+     * will be set once per surf.ctx and per wlStreamResource
+     * if it is already set raise an assertion */
+    assert(surface->ctx.wlStreamResCbQueue == NULL);
+
     queue = wlGetEventQueue(display);
     if (!queue) {
         err = EGL_BAD_ALLOC;
@@ -899,12 +970,13 @@ EGLBoolean wlEglInitializeSurfaceExport(WlEglSurface *surface)
 
     wlExternalApiLock();
     wl_list_init(&surface->oldCtxList);
-    wl_list_insert(&wlEglSurfaceList, &surface->link);
     if (create_surface_context(surface) != EGL_SUCCESS) {
         wlExternalApiUnlock();
         return EGL_FALSE;
     }
 
+    wl_list_insert(&wlEglSurfaceList, &surface->link);
+
     wl_eglstream_display_swap_interval(display->wlStreamDpy,
                                        surface->ctx.wlStreamResource,
                                        surface->swapInterval);
@@ -945,6 +1017,7 @@ resize_callback(struct wl_egl_window *window, void *data)
 
         discard_surface_context(surface);
         surface->ctx.wlStreamResource = NULL;
+        surface->ctx.wlStreamResCbQueue = NULL;
         surface->ctx.isAttached = EGL_FALSE;
         surface->ctx.eglSurface = EGL_NO_SURFACE;
         surface->ctx.eglStream = EGL_NO_STREAM_KHR;
@@ -1420,6 +1493,7 @@ EGLBoolean wlEglQueryNativeResourceHook(EGLDisplay dpy,
     struct wl_eglstream_display *wlStreamDpy = NULL;
     struct wl_eglstream         *wlStream    = NULL;
     EGLBoolean                   res         = EGL_FALSE;
+    EGLint                       originY;
 
     wlExternalApiLock();
 
@@ -1445,11 +1519,21 @@ EGLBoolean wlEglQueryNativeResourceHook(EGLDisplay dpy,
         res = EGL_TRUE;
         goto done;
     case EGL_WAYLAND_Y_INVERTED_WL:
-        /* GLTexture consumers are the only mechanism currently used to map
-         * buffers composited by the wayland compositor. They define the buffer
-         * origin as the lower left corner, which matches to what the wayland
-         * compositor would consider as non-y-inverted */
-        *value = (int)EGL_FALSE;
+        if (wlStreamDpy->exts.stream_origin &&
+            wlStreamDpy->data->egl.queryStream(wlStreamDpy->eglDisplay,
+                                               wlStream->eglStream,
+                                               EGL_STREAM_FRAME_ORIGIN_Y_NV,
+                                               &originY)) {
+            /* If we have an image with origin at the top, the wayland
+             * compositor will consider it as y-inverted */
+            *value = (int)((originY == EGL_TOP_NV) ? EGL_TRUE : EGL_FALSE);
+        } else {
+            /* No mechanism found to query frame orientation. Assume OpenGL
+             * orientation (image origin at the lower left corner), which
+             * matches to what the wayland compositor would consider as
+             * non-y-inverted */
+            *value = (int)EGL_FALSE;
+        }
         res = EGL_TRUE;
         goto done;
     }
